@@ -1,5 +1,8 @@
+# -*- coding: utf-8 -*-
+
 import json
 import os
+import uuid
 from typing import Literal, Optional
 from urllib.parse import urljoin
 
@@ -10,8 +13,23 @@ from .errors import (
     BadRequestException,
     RateLimitException,
     RecordNotFoundException,
+    ServerErrorException,
     UnauthorizedException,
 )
+from .namespaces.core.authentication import Authentication
+from .namespaces.core.databases import Databases
+from .namespaces.core.invites import Invites
+from .namespaces.core.users import Users
+from .namespaces.core.workspaces import Workspaces
+from .namespaces.workspace.branch import Branch
+from .namespaces.workspace.migrations import Migrations
+from .namespaces.workspace.records import Records
+from .namespaces.workspace.search_and_filter import Search_and_filter
+from .namespaces.workspace.table import Table
+
+# TODO this is a manual task, to keep in sync with pyproject.toml
+# could/should be automated to keep in sync
+__version__ = "0.2.0"
 
 PERSONAL_API_KEY_LOCATION = "~/.config/xata/key"
 DEFAULT_BASE_URL_DOMAIN = "xata.sh"
@@ -23,19 +41,6 @@ ApiKeyLocation = Literal["env", "dotenv", "profile", "parameter"]
 WorkspaceIdLocation = Literal["parameter", "env", "config"]
 
 
-class ServerErrorException(Exception):
-    status_code: int
-    message: str
-
-    def __init__(self, status_code, message):
-        self.status_code = status_code
-        self.message = message
-        super().__init__(message)
-
-    def __str__(self) -> str:
-        return f"Server error: {self.status_code} {self.message}"
-
-
 class XataClient:
     """This is the Xata Client. When initialized, it will attempt to read the relevant
     configuration (API key, workspace ID, database name, branch name) from the following
@@ -60,6 +65,7 @@ class XataClient:
 
     configRead: bool = False
     config = None
+    namespaces = {}  # lazy loading container for the namespaces
 
     def __init__(
         self,
@@ -105,14 +111,33 @@ class XataClient:
         self.branch_name = (
             self.get_branch_name_if_configured() if branch_name is None else branch_name
         )
-        # print (
-        #   f"API key: {self.api_key}, "
-        #   f"location: {self.api_key_location}, "
-        #   f"workspaceId: {self.workspace_id}, "
-        #   f"region: {self.region}, "
-        #   f"dbName: {self.db_name}, "
-        #   f"branchName: {self.branch_name}"
-        # )
+        self.headers = {
+            "authorization": f"Bearer {self.api_key}",
+            "user-agent": f"xataio/xata-py:{__version__}",
+            "x-xata-client-id": str(uuid.uuid4()),
+            "x-xata-session-id": str(uuid.uuid4()),
+            "x-xata-agent": f"client=PY_SDK;version={__version__};",
+        }
+
+    def get_config(self) -> dict:
+        """
+        Get the configuration
+        """
+        return {
+            "apiKey": self.api_key,
+            "apiKeyLocation": self.api_key_location,
+            "workspaceId": self.workspace_id,
+            "region": self.region,
+            "dbName": self.db_name,
+            "branchName": self.branch_name,
+            "version": __version__,
+        }
+
+    def get_headers(self) -> dict:
+        """
+        Get the static headers that are iniatilized on client init.
+        """
+        return self.headers
 
     def get_api_key(self) -> tuple[str, ApiKeyLocation]:
         if os.environ.get("XATA_API_KEY") is not None:
@@ -170,9 +195,14 @@ class XataClient:
         return os.environ.get("XATA_BRANCH")
 
     def request(self, method, urlPath, cp=False, headers={}, expect_codes=[], **kwargs):
-        headers["Authorization"] = f"Bearer {self.api_key}"
+        headers = {
+            **headers,
+            **self.headers,
+        }  # TODO use "|" when client py min version >= 3.9
+
         base_url = self.base_url if not cp else self.control_plane_url
         url = urljoin(base_url, urlPath.lstrip("/"))
+
         resp = requests.request(method, url, headers=headers, **kwargs)
         if resp.status_code > 299:
             if resp.status_code in expect_codes:
@@ -300,7 +330,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                     from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :param columns: A list of column names to return. If not provided, all columns are returned.
         :param filter: A filter expression to apply to the query.
         :param sort: A sort expression to apply to the query.
@@ -333,7 +363,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :param columns: A list of column names to return. If not provided, all columns are returned.
         :param filter: A filter expression to apply to the query.
         :param sort: A sort expression to apply to the query.
@@ -369,7 +399,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: A record as a dictionary or None if it doesn't exist.
         """
         db_name, branch_name = self.db_and_branch_names_from_params(
@@ -400,7 +430,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: The ID of the created record.
         """
 
@@ -439,7 +469,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: The ID of the created or updated record.
         """
         db_name, branch_name = self.db_and_branch_names_from_params(
@@ -468,7 +498,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: The ID of the created or updated record.
         """
         db_name, branch_name = self.db_and_branch_names_from_params(
@@ -501,7 +531,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: The updated record.
         """
         db_name, branch_name = self.db_and_branch_names_from_params(
@@ -536,7 +566,7 @@ class XataClient:
         :param db_name: The name of the database to query. If not provided, the database name
                         from the client obejct is used.
         :param branch_name: The name of the branch to query. If not provided, the branch name
-                            from the client obejct is used.
+                        from the client obejct is used.
         :return: The deleted record.
         """
         db_name, branch_name = self.db_and_branch_names_from_params(
@@ -551,3 +581,185 @@ class XataClient:
         if result.status_code == 404:
             raise RecordNotFoundException(id)
         return result.json()
+
+    def search(
+        self,
+        query: str,
+        query_params: dict = {},
+        db_name: str = None,
+        branch_name: str = None,
+    ) -> Optional[dict]:
+        """This endpoint performs full text search across an entire database branch.
+
+        API docs: https://xata.io/docs/api-reference/db/db_branch_name/search#free-text-search
+
+        :meta public:
+        :param query: search string
+        :param query_params: more granular search criteria, see API docs for options
+        :param db_name: The name of the database to query. If not provided, the database name
+                        from the client obejct is used.
+        :param branch_name: The name of the branch to query. If not provided, the branch name
+                        from the client obejct is used.
+
+        :return set of matching records
+        """
+        db_name, branch_name = self.db_and_branch_names_from_params(
+            db_name, branch_name
+        )
+        query_params["query"] = query.strip()
+        result = self.post(
+            f"/db/{db_name}:{branch_name}/search",
+            json=query_params,
+            expect_codes=[200, 400, 403, 404, 500],
+        )
+
+        if result.status_code == 400:
+            raise BadRequestException(result.status_code, result.json()["message"])
+        return result.json()
+
+    def search_table(
+        self,
+        table_name: str,
+        query: str,
+        query_params: dict = {},
+        db_name: str = None,
+        branch_name: str = None,
+    ) -> Optional[dict]:
+        """Run a free text search operation in a particular table.
+
+        :meta public:
+        :param table_name: table to search
+        :param query: search string
+        :param query_params: more granular search criteria, see API docs for options
+        :param db_name: The name of the database to query. If not provided, the database name
+                        from the client obejct is used.
+        :param branch_name: The name of the branch to query. If not provided, the branch name
+                        from the client obejct is used.
+
+        :return set of matching records
+        """
+        db_name, branch_name = self.db_and_branch_names_from_params(
+            db_name, branch_name
+        )
+        query_params["query"] = query.strip()
+        result = self.post(
+            f"/db/{db_name}:{branch_name}/tables/{table_name}/search",
+            json=query_params,
+            expect_codes=[200, 400, 403, 404, 500],
+        )
+
+        if result.status_code == 400:
+            raise BadRequestException(result.status_code, result.json()["message"])
+        return result.json()
+
+    # --------------------------------------------------- #
+    #
+    # Namespace: CORE
+    #
+    # --------------------------------------------------- #
+
+    def authentication(self) -> Authentication:
+        """
+        Authentication Namespace
+        scope: core
+        :return Authentication
+        """
+        if "authentication" not in self.namespaces:
+            self.namespaces["authentication"] = Authentication(self)
+        return self.namespaces["authentication"]
+
+    def databases(self) -> Databases:
+        """
+        Databases Namespace
+        scope: core
+        :return Databases
+        """
+        if "databases" not in self.namespaces:
+            self.namespaces["databases"] = Databases(self)
+        return self.namespaces["databases"]
+
+    def invites(self) -> Invites:
+        """
+        Invites Namespace
+        scope: core
+        :return Invites
+        """
+        if "invites" not in self.namespaces:
+            self.namespaces["invites"] = Invites(self)
+        return self.namespaces["invites"]
+
+    def users(self) -> Users:
+        """
+        Users Namespace
+        scope: core
+        :return Users
+        """
+        if "users" not in self.namespaces:
+            self.namespaces["users"] = Users(self)
+        return self.namespaces["users"]
+
+    def workspaces(self) -> Workspaces:
+        """
+        Workspaces Namespace
+        scope: core
+        :return Workspaces
+        """
+        if "workspaces" not in self.namespaces:
+            self.namespaces["workspaces"] = Workspaces(self)
+        return self.namespaces["workspaces"]
+
+    # --------------------------------------------------- #
+    #
+    # Namespace: WORKSPACE
+    #
+    # --------------------------------------------------- #
+
+    def branch(self) -> Branch:
+        """
+        Branch Namespace
+        scope: workspace
+        :return Branch
+        """
+        if "branch" not in self.namespaces:
+            self.namespaces["branch"] = Branch(self)
+        return self.namespaces["branch"]
+
+    def migrations(self) -> Migrations:
+        """
+        Migrations Namespace
+        scope: workspace
+        :return Migrations
+        """
+        if "migrations" not in self.namespaces:
+            self.namespaces["migrations"] = Migrations(self)
+        return self.namespaces["migrations"]
+
+    def records(self) -> Records:
+        """
+        Records Namespace
+        scope: workspace
+        :return Records
+        """
+        if "records" not in self.namespaces:
+            self.namespaces["records"] = Records(self)
+        return self.namespaces["records"]
+
+    def search_and_filter(self) -> Search_and_filter:
+        """
+        Search_and_Filter Namespace
+        scope: workspace
+        :return Search_and_filter
+        """
+        if "search_and_filter" not in self.namespaces:
+            self.namespaces["search_and_filter"] = Search_and_filter(self)
+        return self.namespaces["search_and_filter"]
+
+    def table(self) -> Table:
+        """
+        Table Namespace
+        scope: workspace
+        :return Table
+        """
+        if "table" not in self.namespaces:
+            self.namespaces["table"] = Table(self)
+        return self.namespaces["table"]
